@@ -12,11 +12,11 @@
 #   Craig Barratt  <cbarratt@users.sourceforge.net>
 #
 # COPYRIGHT
-#   Copyright (C) 2004-2013  Craig Barratt
+#   Copyright (C) 2004-2020  Craig Barratt
 #
-#   This program is free software; you can redistribute it and/or modify
+#   This program is free software: you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
-#   the Free Software Foundation; either version 2 of the License, or
+#   the Free Software Foundation, either version 3 of the License, or
 #   (at your option) any later version.
 #
 #   This program is distributed in the hope that it will be useful,
@@ -25,12 +25,11 @@
 #   GNU General Public License for more details.
 #
 #   You should have received a copy of the GNU General Public License
-#   along with this program; if not, write to the Free Software
-#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 #========================================================================
 #
-# Version 3.3.0, released 14 Apr 2013.
+# Version 4.4.0, released 20 Jun 2020.
 #
 # See http://backuppc.sourceforge.net.
 #
@@ -39,20 +38,18 @@
 package BackupPC::Storage::Text;
 
 use strict;
-use vars qw(%Conf);
+use vars qw(%Conf %Status %Info);
 use Data::Dumper;
 use File::Path;
 use Fcntl qw/:flock/;
+use Storable qw(store retrieve fd_retrieve store_fd);
 
 sub new
 {
     my $class = shift;
     my($flds, $paths) = @_;
 
-    my $s = bless {
-	%$flds,
-	%$paths,
-    }, $class;
+    my $s = bless {%$flds, %$paths}, $class;
     return $s;
 }
 
@@ -69,32 +66,50 @@ sub setPaths
 sub BackupInfoRead
 {
     my($s, $host) = @_;
-    local(*BK_INFO, *LOCK);
-    my(@Backups);
+    my(@Backups, $bkFd, $lockFd, $locked);
 
-    flock(LOCK, LOCK_EX) if open(LOCK, "$s->{TopDir}/pc/$host/LOCK");
-    if ( open(BK_INFO, "$s->{TopDir}/pc/$host/backups") ) {
-	binmode(BK_INFO);
-        while ( <BK_INFO> ) {
+    if ( open($lockFd, ">", "$s->{TopDir}/pc/$host/LOCK") ) {
+        flock($lockFd, LOCK_EX);
+        $locked = 1;
+    }
+    if ( open($bkFd, "$s->{TopDir}/pc/$host/backups") ) {
+        binmode($bkFd);
+        while ( <$bkFd> ) {
             s/[\n\r]+//;
-            next if ( !/^(\d+\t(incr|full|partial).*)/ );
+            next if ( !/^(\d+\t(incr|full|partial|active).*)/ );
             $_ = $1;
             @{$Backups[@Backups]}{@{$s->{BackupFields}}} = split(/\t/);
         }
-        close(BK_INFO);
+        close($bkFd);
+    }
+    if ( $locked ) {
+        flock($lockFd, LOCK_UN);
+        close($lockFd);
     }
-    close(LOCK);
-    #
-    # Default the version field.  Prior to 3.0.0 the xferMethod
-    # field is empty, so we use that to figure out the version.
-    #
     for ( my $i = 0 ; $i < @Backups ; $i++ ) {
-        next if ( $Backups[$i]{version} ne "" );
-        if ( $Backups[$i]{xferMethod} eq "" ) {
-            $Backups[$i]{version} = "2.1.2";
-        } else {
-            $Backups[$i]{version} = "3.0.0";
+        #
+        # Default the version field.  Prior to 3.0.0 the xferMethod
+        # field is empty, so we use that to figure out the version.
+        #
+        if ( $Backups[$i]{version} eq "" ) {
+            if ( $Backups[$i]{xferMethod} eq "" ) {
+                $Backups[$i]{version} = "2.1.2";
+            } else {
+                $Backups[$i]{version} = "3.0.0";
+            }
         }
+        #
+        # Reconstitue share2path into a hash
+        #
+        if ( $Backups[$i]{share2path} ne "" ) {
+            my $str = $Backups[$i]{share2path};
+            $str =~ s{%(..)}{chr(hex($1))}eg;
+            $Backups[$i]{share2path} = eval $str;
+            if ( $@ ) {
+                print(STDERR "BackupInfoRead: host $host: can't rebuild share2path from $str\n");
+            }
+        }
+        $Backups[$i]{comment} =~ s{%(..)}{chr(hex($1))}eg;
     }
     return @Backups;
 }
@@ -102,16 +117,26 @@ sub BackupInfoRead
 sub BackupInfoWrite
 {
     my($s, $host, @Backups) = @_;
-    my($i, $contents, $fileOk);
+    my($i, $contents);
 
     #
     # Generate the file contents
     #
     for ( $i = 0 ; $i < @Backups ; $i++ ) {
         my %b = %{$Backups[$i]};
+        if ( ref($b{share2path}) eq 'HASH' ) {
+            my $dump = Data::Dumper->new([$b{share2path}]);
+            $dump->Indent(0);
+            $dump->Sortkeys(1);
+            $dump->Terse(1);
+            my $value = $dump->Dump;
+            $value =~ s{([%\t\n\r])}{sprintf("%%%02x", ord($1))}eg;
+            $b{share2path} = $value;
+        }
+        $b{comment} =~ s{([%\t\n\r])}{sprintf("%%%02x", ord($1))}eg;
         $contents .= join("\t", @b{@{$s->{BackupFields}}}) . "\n";
     }
-    
+
     #
     # Write the file
     #
@@ -121,29 +146,33 @@ sub BackupInfoWrite
 sub RestoreInfoRead
 {
     my($s, $host) = @_;
-    local(*RESTORE_INFO, *LOCK);
-    my(@Restores);
+    my(@Restores, $resFd, $lockFd, $locked);
 
-    flock(LOCK, LOCK_EX) if open(LOCK, "$s->{TopDir}/pc/$host/LOCK");
-    if ( open(RESTORE_INFO, "$s->{TopDir}/pc/$host/restores") ) {
-	binmode(RESTORE_INFO);
-        while ( <RESTORE_INFO> ) {
+    if ( open($lockFd, ">", "$s->{TopDir}/pc/$host/LOCK") ) {
+        flock($lockFd, LOCK_EX);
+        $locked = 1;
+    }
+    if ( open($resFd, "$s->{TopDir}/pc/$host/restores") ) {
+        binmode($resFd);
+        while ( <$resFd> ) {
             s/[\n\r]+//;
             next if ( !/^(\d+.*)/ );
             $_ = $1;
             @{$Restores[@Restores]}{@{$s->{RestoreFields}}} = split(/\t/);
         }
-        close(RESTORE_INFO);
+        close($resFd);
+    }
+    if ( $locked ) {
+        flock($lockFd, LOCK_UN);
+        close($lockFd);
     }
-    close(LOCK);
     return @Restores;
 }
 
 sub RestoreInfoWrite
 {
     my($s, $host, @Restores) = @_;
-    local(*RESTORE_INFO, *LOCK);
-    my($i, $contents, $fileOk);
+    my($i, $contents);
 
     #
     # Generate the file contents
@@ -162,29 +191,33 @@ sub RestoreInfoWrite
 sub ArchiveInfoRead
 {
     my($s, $host) = @_;
-    local(*ARCHIVE_INFO, *LOCK);
-    my(@Archives);
+    my(@Archives, $archFd, $lockFd, $locked);
 
-    flock(LOCK, LOCK_EX) if open(LOCK, "$s->{TopDir}/pc/$host/LOCK");
-    if ( open(ARCHIVE_INFO, "$s->{TopDir}/pc/$host/archives") ) {
-        binmode(ARCHIVE_INFO);
-        while ( <ARCHIVE_INFO> ) {
+    if ( open($lockFd, ">", "$s->{TopDir}/pc/$host/LOCK") ) {
+        flock($lockFd, LOCK_EX);
+        $locked = 1;
+    }
+    if ( open($archFd, "$s->{TopDir}/pc/$host/archives") ) {
+        binmode($archFd);
+        while ( <$archFd> ) {
             s/[\n\r]+//;
             next if ( !/^(\d+.*)/ );
             $_ = $1;
             @{$Archives[@Archives]}{@{$s->{ArchiveFields}}} = split(/\t/);
         }
-        close(ARCHIVE_INFO);
+        close($archFd);
+    }
+    if ( $locked ) {
+        flock($lockFd, LOCK_UN);
+        close($lockFd);
     }
-    close(LOCK);
     return @Archives;
 }
 
 sub ArchiveInfoWrite
 {
     my($s, $host, @Archives) = @_;
-    local(*ARCHIVE_INFO, *LOCK);
-    my($i, $contents, $fileOk);
+    my($i, $contents);
 
     #
     # Generate the file contents
@@ -209,8 +242,7 @@ sub ArchiveInfoWrite
 sub TextFileWrite
 {
     my($s, $file, $contents) = @_;
-    local(*FD, *LOCK);
-    my($fileOk);
+    my($fileOk, $fd);
 
     (my $dir = $file) =~ s{(.+)/(.+)}{$1};
 
@@ -218,38 +250,47 @@ sub TextFileWrite
         eval { mkpath($dir, 0, 0775) };
         return "TextFileWrite: can't create directory $dir" if ( $@ );
     }
-    if ( open(FD, ">", "$file.new") ) {
-	binmode(FD);
-        print FD $contents;
-        close(FD);
+    if ( open($fd, ">", "$file.new") ) {
+        binmode($fd);
+        print $fd $contents;
+        close($fd);
         #
         # verify the file
         #
-        if ( open(FD, "<", "$file.new") ) {
-            binmode(FD);
-            if ( join("", <FD>) ne $contents ) {
+        if ( open($fd, "<", "$file.new") ) {
+            binmode($fd);
+            if ( join("", <$fd>) ne $contents ) {
                 return "TextFileWrite: Failed to verify $file.new";
             } else {
                 $fileOk = 1;
             }
-            close(FD);
+            close($fd);
         }
+        #
+        # use same gid as original file
+        #
+        my $uid = (stat("$file.new"))[4];
+        my $gid = (stat($file))[5];
+        chown($uid, $gid, "$file.new") if ( defined($uid) && defined($gid) );
     }
     if ( $fileOk ) {
-        my $lock;
-        
-        if ( open(LOCK, "$dir/LOCK") || open(LOCK, ">", "$dir/LOCK") ) {
-            $lock = 1;
-            flock(LOCK, LOCK_EX);
+        my($locked, $lockFd);
+
+        if ( open($lockFd, ">", "$dir/LOCK") ) {
+            $locked = 1;
+            flock($lockFd, LOCK_EX);
         }
         if ( -s "$file" ) {
-            unlink("$file.old")           if ( -f "$file.old" );
-            rename("$file", "$file.old")  if ( -f "$file" );
+            unlink("$file.old")          if ( -f "$file.old" );
+            rename("$file", "$file.old") if ( -f "$file" );
         } else {
             unlink("$file") if ( -f "$file" );
         }
         rename("$file.new", "$file") if ( -f "$file.new" );
-        close(LOCK) if ( $lock );
+        if ( $locked ) {
+            flock($lockFd, LOCK_UN);
+            close($lockFd);
+        }
     } else {
         return "TextFileWrite: Failed to write $file.new";
     }
@@ -265,9 +306,9 @@ sub ConfigPath
         return "$s->{ConfDir}/pc/$host.pl";
     } else {
         return "$s->{TopDir}/pc/$host/config.pl"
-            if ( -f "$s->{TopDir}/pc/$host/config.pl" );
+          if ( -f "$s->{TopDir}/pc/$host/config.pl" );
         return "$s->{ConfDir}/$host.pl"
-            if ( $host ne "config" && -f "$s->{ConfDir}/$host.pl" );
+          if ( $host ne "config" && -f "$s->{ConfDir}/$host.pl" );
         return "$s->{ConfDir}/pc/$host.pl";
     }
 }
@@ -280,14 +321,14 @@ sub ConfigDataRead
     #
     # TODO: add lock
     #
-    my $conf = $prevConfig || {};
+    my $conf       = $prevConfig || {};
     my $configPath = $s->ConfigPath($host);
 
     push(@configs, $configPath) if ( -f $configPath );
     foreach $config ( @configs ) {
         %Conf = %$conf;
         if ( !defined($ret = do $config) && ($! || $@) ) {
-            $mesg = "Couldn't open $config: $!" if ( $! );
+            $mesg = "Couldn't open $config: $!"    if ( $! );
             $mesg = "Couldn't execute $config: $@" if ( $@ );
             $mesg =~ s/[\n\r]+//;
             return ($mesg, $conf);
@@ -300,9 +341,9 @@ sub ConfigDataRead
     #
     foreach my $param ( qw(BackupFilesOnly BackupFilesExclude) ) {
         next if ( !defined($conf->{$param}) || ref($conf->{$param}) eq "HASH" );
-        $conf->{$param} = [ $conf->{$param} ]
-                                if ( ref($conf->{$param}) ne "ARRAY" );
-        $conf->{$param} = { "*" => $conf->{$param} };
+        $conf->{$param} = [$conf->{$param}]
+          if ( ref($conf->{$param}) ne "ARRAY" );
+        $conf->{$param} = {"*" => $conf->{$param}};
     }
 
     #
@@ -310,18 +351,38 @@ sub ConfigDataRead
     # BlackoutHourEnd, and BlackoutWeekDays parameters.
     #
     if ( defined($conf->{BlackoutHourBegin}) ) {
-        push(@{$conf->{BlackoutPeriods}},
-             {
-                 hourBegin => $conf->{BlackoutHourBegin},
-                 hourEnd   => $conf->{BlackoutHourEnd},
-                 weekDays  => $conf->{BlackoutWeekDays},
-             }
+        push(
+            @{$conf->{BlackoutPeriods}},
+            {
+                hourBegin => $conf->{BlackoutHourBegin},
+                hourEnd   => $conf->{BlackoutHourEnd},
+                weekDays  => $conf->{BlackoutWeekDays},
+            }
         );
         delete($conf->{BlackoutHourBegin});
         delete($conf->{BlackoutHourEnd});
         delete($conf->{BlackoutWeekDays});
     }
 
+    #
+    # Check that certain settings have valid values
+    #
+    if (   $conf->{BackupPCNightlyPeriod} != 1
+        && $conf->{BackupPCNightlyPeriod} != 2
+        && $conf->{BackupPCNightlyPeriod} != 4
+        && $conf->{BackupPCNightlyPeriod} != 8
+        && $conf->{BackupPCNightlyPeriod} != 16 ) {
+        $conf->{BackupPCNightlyPeriod} = 1;
+    }
+    if (   $conf->{PoolSizeNightlyUpdatePeriod} != 0
+        && $conf->{PoolSizeNightlyUpdatePeriod} != 1
+        && $conf->{PoolSizeNightlyUpdatePeriod} != 2
+        && $conf->{PoolSizeNightlyUpdatePeriod} != 4
+        && $conf->{PoolSizeNightlyUpdatePeriod} != 8
+        && $conf->{PoolSizeNightlyUpdatePeriod} != 16 ) {
+        $conf->{PoolSizeNightlyUpdatePeriod} = 16;
+    }
+
     return (undef, $conf);
 }
 
@@ -345,19 +406,18 @@ sub ConfigDataWrite
 sub ConfigFileMerge
 {
     my($s, $inFile, $newConf) = @_;
-    local(*C);
-    my($contents, $skipExpr, $fakeVar);
+    my($contents, $skipExpr, $fakeVar, $configFd);
     my $done = {};
 
     if ( -f $inFile ) {
         #
         # Match existing settings in current config file
         #
-        open(C, $inFile)
-            || return ("ConfigFileMerge: can't open/read $inFile", undef);
-        binmode(C);
+        open($configFd, $inFile)
+          || return ("ConfigFileMerge: can't open/read $inFile", undef);
+        binmode($configFd);
 
-        while ( <C> ) {
+        while ( <$configFd> ) {
             if ( /^\s*\$Conf\{([^}]*)\}\s*=(.*)/ ) {
                 my $var = $1;
                 $skipExpr = "\$fakeVar = $2\n";
@@ -365,6 +425,7 @@ sub ConfigFileMerge
                     my $d = Data::Dumper->new([$newConf->{$var}], [*value]);
                     $d->Indent(1);
                     $d->Terse(1);
+                    $d->Sortkeys(1);
                     my $value = $d->Dump;
                     $value =~ s/(.*)\n/$1;\n/s;
                     $contents .= "\$Conf{$var} = " . $value;
@@ -375,9 +436,7 @@ sub ConfigFileMerge
             } else {
                 $contents .= $_;
             }
-            if ( defined($skipExpr)
-                    && ($skipExpr =~ /^\$fakeVar = *<</
-                        || $skipExpr =~ /;[\n\r]*$/) ) {
+            if ( defined($skipExpr) && ($skipExpr =~ /^\$fakeVar = *<</ || $skipExpr =~ /;[\n\r]*$/) ) {
                 #
                 # if we have a complete expression, then we are done
                 # skipping text from the original config file.
@@ -387,21 +446,22 @@ sub ConfigFileMerge
                 $skipExpr = undef if ( $@ eq "" );
             }
         }
-        close(C);
+        close($configFd);
     }
 
     #
     # Add new entries not matched in current config file
     #
     foreach my $var ( sort(keys(%$newConf)) ) {
-	next if ( $done->{$var} );
-	my $d = Data::Dumper->new([$newConf->{$var}], [*value]);
-	$d->Indent(1);
-	$d->Terse(1);
-	my $value = $d->Dump;
-	$value =~ s/(.*)\n/$1;\n/s;
-	$contents .= "\$Conf{$var} = " . $value;
-	$done->{$var} = 1;
+        next if ( $done->{$var} );
+        my $d = Data::Dumper->new([$newConf->{$var}], [*value]);
+        $d->Indent(1);
+        $d->Terse(1);
+        $d->Sortkeys(1);
+        my $value = $d->Dump;
+        $value =~ s/(.*)\n/$1;\n/s;
+        $contents .= "\$Conf{$var} = " . $value;
+        $done->{$var} = 1;
     }
     return (undef, $contents);
 }
@@ -415,6 +475,35 @@ sub ConfigMTime
     return (stat($s->ConfigPath()))[9];
 }
 
+sub StatusDataRead
+{
+    my($s) = @_;
+    my($ret, $mesg);
+
+    %Status = ();
+    %Info   = ();
+    if (   -f "$s->{LogDir}/status.pl"
+        && !defined($ret = do "$s->{LogDir}/status.pl")
+        && ($! || $@) ) {
+        $mesg = "Couldn't open $s->{LogDir}/status.pl: $!"    if ( $! );
+        $mesg = "Couldn't execute $s->{LogDir}/status.pl: $@" if ( $@ );
+        $mesg =~ s/[\n\r]+//;
+        rename("$s->{LogDir}/status.pl", "$s->{LogDir}/status.pl.bad");
+        return ($mesg, undef);
+    }
+    return (\%Status, \%Info);
+}
+
+sub StatusDataWrite
+{
+    my($s, $status, $info) = @_;
+
+    my($dump) = Data::Dumper->new([$info, $status], [qw(*Info *Status)]);
+    $dump->Indent(1);
+    my $text = $dump->Dump;
+    $s->TextFileWrite("$s->{LogDir}/status.pl", $text);
+}
+
 #
 # Returns information from the host file in $s->{ConfDir}/hosts.
 # With no argument a ref to a hash of hosts is returned.  Each
@@ -425,17 +514,23 @@ sub ConfigMTime
 sub HostInfoRead
 {
     my($s, $host) = @_;
-    my(%hosts, @hdr, @fld);
-    local(*HOST_INFO, *LOCK);
+    my(%hosts, @hdr, @fld, $hostFd, $lockFd, $locked);
+    my(@Backups, $bkFd);
 
-    flock(LOCK, LOCK_EX) if open(LOCK, "$s->{ConfDir}/LOCK");
-    if ( !open(HOST_INFO, "$s->{ConfDir}/hosts") ) {
+    if ( open($lockFd, ">", "$s->{ConfDir}/LOCK") ) {
+        flock($lockFd, LOCK_EX);
+        $locked = 1;
+    }
+    if ( !open($hostFd, "$s->{ConfDir}/hosts") ) {
         print(STDERR "Can't open $s->{ConfDir}/hosts\n");
-        close(LOCK);
+        if ( $locked ) {
+            flock($lockFd, LOCK_UN);
+            close($lockFd);
+        }
         return {};
     }
-    binmode(HOST_INFO);
-    while ( <HOST_INFO> ) {
+    binmode($hostFd);
+    while ( <$hostFd> ) {
         s/[\n\r]+//;
         s/#.*//;
         s/\s+$//;
@@ -443,7 +538,7 @@ sub HostInfoRead
         #
         # Split on white space, except if preceded by \
         # using zero-width negative look-behind assertion
-	# (always wanted to use one of those).
+        # (always wanted to use one of those).
         #
         @fld = split(/(?<!\\)\s+/, $1);
         #
@@ -456,8 +551,11 @@ sub HostInfoRead
             if ( defined($host) ) {
                 next if ( lc($fld[0]) ne lc($host) );
                 @{$hosts{lc($fld[0])}}{@hdr} = @fld;
-		close(HOST_INFO);
-                close(LOCK);
+                close($hostFd);
+                if ( $locked ) {
+                    flock($lockFd, LOCK_UN);
+                    close($lockFd);
+                }
                 return \%hosts;
             } else {
                 @{$hosts{lc($fld[0])}}{@hdr} = @fld;
@@ -466,8 +564,11 @@ sub HostInfoRead
             @hdr = @fld;
         }
     }
-    close(HOST_INFO);
-    close(LOCK);
+    close($hostFd);
+    if ( $locked ) {
+        flock($lockFd, LOCK_UN);
+        close($lockFd);
+    }
     return \%hosts;
 }
 
@@ -481,23 +582,20 @@ sub HostInfoRead
 sub HostInfoWrite
 {
     my($s, $hosts) = @_;
-    my($gotHdr, @fld, $hostText, $contents);
-    local(*HOST_INFO);
+    my($gotHdr, @fld, $hostText, $contents, $hostFd);
 
-    if ( !open(HOST_INFO, "$s->{ConfDir}/hosts") ) {
+    if ( !open($hostFd, "$s->{ConfDir}/hosts") ) {
         return "Can't open $s->{ConfDir}/hosts";
     }
     foreach my $host ( keys(%$hosts) ) {
         my $name = "$hosts->{$host}{host}";
-        my $rest = "\t$hosts->{$host}{dhcp}"
-                 . "\t$hosts->{$host}{user}"
-                 . "\t$hosts->{$host}{moreUsers}";
+        my $rest = "\t$hosts->{$host}{dhcp}\t$hosts->{$host}{user}\t$hosts->{$host}{moreUsers}";
         $name =~ s/ /\\ /g;
         $rest =~ s/ //g;
         $hostText->{$host} = $name . $rest;
     }
-    binmode(HOST_INFO);
-    while ( <HOST_INFO> ) {
+    binmode($hostFd);
+    while ( <$hostFd> ) {
         s/[\n\r]+//;
         if ( /^\s*$/ || /^\s*#/ ) {
             $contents .= $_ . "\n";
@@ -524,7 +622,7 @@ sub HostInfoWrite
         $contents .= $hostText->{$host} . "\n";
         delete($hostText->{$host});
     }
-    close(HOST_INFO);
+    close($hostFd);
 
     #
     # Write and verify the new host file
